#include <stdlib.h>
#include <string.h>
#include "defs.h"
#include "funclets.h"
#include "HIL.h"
#include "JTAGfunc.h"
#include "MSP430mspgcc.h"

#ifndef min
    #define min(a,b) (((a)<(b)) ? (a) : (b))
#endif

// ----------------------------------------------------------------------------
/* msp430 code and buffer to write flash

As it's not implemented to read the MSP430 registers, return values have to be
passed in memory. Due to a JTAG bug in certain devices that can corrupt RAM, the
return value is passed in CCR0.
*/
#include "progFlash.ci"         //include the flash program

WORD ramsize = 256;             //TODO: get device memory size
void (*flash_callback)(WORD, WORD) = NULL;

STATUS_T programFlash(WORD address, const CHAR* buffer, WORD count) {
    STATUS_T res = STATUS_OK;
    WORD blocksize;
    WORD done = 0;
    WORD tries;
    WORD *code = malloc(ramsize);
    if (code == NULL)   return STATUS_ERROR;                            //Out of memory
        
    MSP430_Log(2, "funclets: Flash write...\n");
    memcpy(code, funclet_progFlash, sizeof(funclet_progFlash));         //copy program to working buffer
    while (done < count) {                                              //download in blocks
        blocksize = min(count-done, ramsize-sizeof(funclet_progFlash)); //calculate blocksize
        memcpy(&code[sizeof(funclet_progFlash)/sizeof(WORD)], &buffer[done], blocksize); //copy new data to be flashed
        //fill in arguments
        code[3] = 0;//srinit;                                           //TODO: need arg for DCO+?
        code[4] = address;                                              //starting address in flash
        code[5] = blocksize/2;                                          //number of words to write
        MSP430_Log(2, "funclets: Flash write at 0x%04x %d bytes\n", address, blocksize);
        //download data and start flash programing funclet in RAM
        for (tries=3; tries; tries--) {
            res =  executeCode(code, (sizeof(funclet_progFlash) + blocksize)/sizeof(WORD), 1, 1);
            if (res != STATUS_OK) {
                MSP430_Log(1, "funclets: Flash write retrying in block 0x%04x\n", address);
                GetDevice();
            }
        }
        if (res != STATUS_OK) break;
        if (ReadMem(F_WORD, CCR0_ADDR)) {                               //read CCR0, it's used as return value
            MSP430_Log(1, "funclets: Flash write error in block 0x%04x\n", address);
            res = STATUS_ERROR;
            break;
        }
        //~ MSP430_Log(2, "block 0x%04x OK\n", address);
        address += blocksize;                                           //adjust for next block
        done += blocksize;                                              //advance in source buffer
        if (flash_callback != NULL) {
            flash_callback(done, count);
        }
    }
    free(code);                                                         //clean up memory
    MSP430_Log(2, "funclets: Flash write finished\n");
    return res;
}

// ----------------------------------------------------------------------------
/* msp430 code to erase flash
*/
#include "eraseFlash.ci"                        //include the program to erase the flash

STATUS_T eraseFlash(WORD type, WORD address) {
    WORD fctl1;
    //select correct erase mode
    switch (type) {
        case ERASE_SEGMENT:     fctl1 = 0xA502; break;
        case ERASE_MAIN:        fctl1 = 0xA504; break;
        case ERASE_ALL:         fctl1 = 0xA506; break;
        default: return (STATUS_ERROR);
    }
    //fill in arguments
    funclet_eraseFlash[3] = 0;//srinit;         //TODO: need arg for DCO+?
    funclet_eraseFlash[4] = fctl1;              //set FCTL1 contents
    funclet_eraseFlash[5] = address;            //address within the segment to be erased
    //download erase prog and execute
    return executeCode(funclet_eraseFlash, sizeof(funclet_eraseFlash)/sizeof(WORD), 1, 1);
}

// ----------------------------------------------------------------------------

//The following functions were originaly developed by TI. I added debug outputs an
//simpified some parts
//questions to chris.

/* Read Memory Address Bus.
This function can be used at any time (when the JTAG is syncronized)
I may however return bogus results while the CPU is running.
Using it in a loop and multiple checks if it returns the same address
is a probable mean to check if the CPU waits on a "jmp $" insn.
*/
static WORD readMab(void) {
    HIL_JTAG_IR(IR_ADDR_CAPTURE);
    return ((WORD) HIL_JTAG_DR(0, 16));
}

/*wait until CPU is on an instruction load boundary*/
STATUS_T syncCPU(void) {
    WORD cntrlSig;
    int i;
    
    HIL_JTAG_IR(IR_CNTRL_SIG_CAPTURE);
    cntrlSig = (WORD) HIL_JTAG_DR(0, F_WORD);
    
    HIL_TCLK(1);
    
    if (!(cntrlSig & CNTRL_SIG_TCE)) {                  //check if sync is lost
        HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);                //Lch
        HIL_JTAG_DR(0x2401, F_WORD);                    // Set device into JTAG mode + read
        //~ //HIL_JTAG_DR(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU | (cntrlSig & 0xff), F_WORD);
        
        for (i = 50; i > 0; i--) {
            cntrlSig = HIL_JTAG_DR(0x0000, F_WORD);
            MSP430_Log(6, "funclets: JTAG CNTRL: 0x%04x\n", cntrlSig);
            if (cntrlSig & CNTRL_SIG_TCE) {
                MSP430_Log(5, "funclets: Sync OK\n");   //DEBUG
                break;
            } else if (i == 1) {
                MSP430_Log(5, "funclets: Sync failed\n"); //DEBUG
                return(STATUS_ERROR);                   // Timeout reached, return false
            } else {
                HIL_DelayMSec(1);
            }
        }
    }
    
    if (cntrlSig & CNTRL_SIG_CPU_HALT) {
        HIL_TCLK(0);
        // Clear HALT. Read/Write is under JTAG control. As a precaution, disable interrupts.
        HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
        HIL_JTAG_DR(CNTRL_SIG_READ|CNTRL_SIG_TCE1|CNTRL_SIG_TAGFUNCSAT, 16);
        HIL_TCLK(1);
    } else {  
        // Read/Write is under JTAG control. As a precaution, disable interrupts.
        HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
        HIL_JTAG_DR(CNTRL_SIG_READ|CNTRL_SIG_TCE1|CNTRL_SIG_TAGFUNCSAT, 16);
    }

    // Advance to an instruction load boundary. Normal the device will be on an instruction load boundary
    // following the synchronization. However, a bug in the F413 can stop the CPU off of an instruction
    // load boundary. In that case, manually advance the CPU to an instruction load boundary. Unfortunately,
    // at this time the device type can be unknown. Consequently, advance to an instruction load boundary
    // for all devices.
    if (SetInstrFetch() != STATUS_OK) {
        MSP430_Log(1, "funclets: Not in Fetch\n"); //DEBUG
        return (STATUS_ERROR); // Synchronization failed!
    }
    
    // Execute a dummy instruction (MOV R3,R3) to work around a problem in the F12x that can sometimes cause
    // the first TCLK after synchronization to be lost.
    // Note: It's critical that the dummy instruction require only a single cycle.
    // Note: Since the state of the PC is not critical at this time, there is no need to restore/adjust it.
    // Note: It is known that at this time that interrupts are not possible (GIE is clear).
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(0x4303, F_WORD);
    HIL_TCLK(POS_EDGE);
    
    // (As in internal check) Must exit on an instruction load boundary.
    HIL_JTAG_IR(IR_CNTRL_SIG_CAPTURE);
    if (!(HIL_JTAG_DR(0, F_WORD) & CNTRL_SIG_INSTRLOAD)) {
        MSP430_Log(1, "funclets: Not in Fetch\n"); //DEBUG
        return (STATUS_ERROR);
    } else {
        return (STATUS_OK);
    }
}

// Set PC and avoid corrupting RAM while doing so. used in executeCode
static STATUS_T setPCsave(WORD addr) {
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(MOV_IMM_PC, F_WORD);            // Force PC into non-RAM area.
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_DR(addr, F_WORD);
    HIL_TCLK(NEG_EDGE);
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_IR(IR_ADDR_CAPTURE);               // Verify the Program Counter.
    if (HIL_JTAG_DR(0, F_WORD) != addr) {
        return (STATUS_ERROR);
    }
    return (STATUS_OK);
}

/* This function downloads programs (funclets) into RAM, and starts their execution.
F413P, F41xC, and F12x devices contain a silicon bug that can corrupt memory when
the JTAG instruction register or the JTAG data register is accessed while TCLK is
high. The following sequence of steps has been carefully crafted to work around
this bug. The bug destroys the contents of a RAM under the following conditions:
- The device is under JTAG control
- MCLK/TCLK is high
- MAB selects a RAM address
- A JTAG instruction is shifted in or a JTAG data is shifted in
*/
STATUS_T executeCode(const WORD* code, WORD sizeCode, BOOL verify, BOOL wait) {
    int i;

    if (sizeCode == 0) {
        return (STATUS_OK);
    }
    if (code[0] & 1) {                          // Address must be even.
        return (STATUS_ERROR);
    }
    if (SetInstrFetch() != STATUS_OK) {         //Must start on an instruction load boundary.
        return (STATUS_ERROR);
    }

    MSP430_Log(3, "funclets: download %d words...\n", sizeCode); //DEBUG
    SetPC(code[0] - 2);
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(JMP_$, F_WORD);                 // Load "jmp $" instruction for use while HALT'ed.
    HIL_TCLK(NEG_EDGE);

    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2408, F_WORD);                // Set JTAG_HALT bit
    HIL_JTAG_IR(IR_DATA_QUICK);
    for (i = 0; i < sizeCode; i++) {
        HIL_JTAG_DR(code[i], F_WORD);           // Write data to memory.
        HIL_TCLK(NEG_EDGE);
    }
    HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
    HIL_JTAG_DR(0x2401, F_WORD);                // Clear the HALT_JTAG bit (Release CPU)
    
    if (verify) {       // Verify the memory contents (without corrupting it!)
        WORD *Read = (WORD *) malloc(sizeCode*sizeof(WORD));
        MSP430_Log(5, "funclets: verify...\n"); //DEBUG
        if (setPCsave(code[0] - 2) != STATUS_OK) {
            free(Read);                         // Free read buffer
            MSP430_Log(1, "funclets: setPCsave failed\n"); //DEBUG
            return (STATUS_ERROR);
        }
        HIL_JTAG_IR(IR_DATA_16BIT);
        HIL_JTAG_DR(JMP_$, F_WORD);             // Load "jmp $" instruction for use while HALT'ed.
        HIL_TCLK(NEG_EDGE);

        HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
        HIL_JTAG_DR(0x2409, F_WORD);            // Set JTAG_HALT bit
        HIL_JTAG_IR(IR_DATA_QUICK);
        for (i = 0; i < sizeCode; i++) {
            HIL_TCLK(NEG_EDGE);
            Read[i] = HIL_JTAG_DR(0, F_WORD);   // Read data from memory.
        }
        HIL_JTAG_IR(IR_CNTRL_SIG_16BIT);
        HIL_JTAG_DR(0x2401, F_WORD);            // Clear the HALT_JTAG bit (Release CPU)

        for (i = 0; i < sizeCode; i++) {        // Verify the memory contents.
            if (Read[i] != code[i]) {
                free(Read);                     // Free read buffer
                MSP430_Log(1, "funclets: vfy error (word %d)\n", i);   //DEBUG
                return (STATUS_ERROR);
            }
        }
        MSP430_Log(5, "funclets: vfy ok\n");    //DEBUG
        free(Read);                             // Free read buffer
    } // if (verify)
    
    //this block is needed if device has JTAG bug, but it shouldn't hurt anyway
    MSP430_Log(5, "funclets: park PC...\n");    //DEBUG
    if (setPCsave(ROM_ADDR) != STATUS_OK) {
        return (STATUS_ERROR);
    }
    
    MSP430_Log(5, "funclets: set active...\n"); //DEBUG
    HIL_JTAG_IR(IR_DATA_16BIT);
    HIL_JTAG_DR(0x4032, 16);                    // Clear the Status Register (no low power modes, diable irqs).
    HIL_TCLK(NEG_EDGE);
    HIL_JTAG_DR(0, 16);
    HIL_TCLK(1);
    HIL_TCLK(POS_EDGE);

    MSP430_Log(5, "funclets: set PC\n");        //DEBUG
    if (setPCsave(code[1]) != STATUS_OK) {      // Load the Program Counter.
        return (STATUS_ERROR);
    }

    MSP430_Log(5, "funclets: run...\n");        //DEBUG
    HIL_JTAG_IR(IR_CNTRL_SIG_RELEASE);          // Release the CPU (but retain JTAG control signals).
    HIL_TCLK(1);

    if (wait) {
        // Limit the number of cycles for the code to execute.
        for (i = 0; i < 20000; i++) {
            // Poll the device until it is executing the final code instruction.
            if ((readMab() == code[2]) && (readMab() == code[2]) && (readMab() == code[2])) {
                // For F413P, F41xC, and F12x devices, the following step will likely corrupt RAM.
                // However, we don't care about this as the RAM contents will be reloaded in necessary.
                MSP430_Log(5, "funclets: wait OK\n"); //DEBUG
                return (syncCPU());             // Resume control of the CPU. Read/Write will be set.
            }
        }
        MSP430_Log(1, "funclets: wait failed\n"); //DEBUG
        return (STATUS_ERROR);
    }
    MSP430_Log(3, "funclets: exec OK\n");       //DEBUG
    return (STATUS_OK);
}

//check if the cpu is executing code or if its halted
//may have some uncertainity if it hits a loop repeatedly on the same instruction...
int isHalted(void)
{
    int i;
    WORD reference = readMab();
    HIL_JTAG_IR(IR_CNTRL_SIG_RELEASE);          // Release the CPU (but retain JTAG control signals).
    HIL_TCLK(1);
    for (i=0; i<10; i++)
    {
        HIL_DelayMSec(1); //make sure that the target has the chance to execute some code
        if (readMab() != reference) {
            return 0;
        }
    }
    return 1;
}
